from base_st_classes import *
import sys

class RebootSuite(NFSSuite):
    """Tests that should be run during the grace period after server reboot

    NOTE these tests should NOT be run with the rest of the test suite.
    Right now, they are intended to be run with the server restarted by hand
    upon demand.
    """

    """
    NOTE - what is the difference between BAD and CONFLICT?  is there a
    clear case where one must be returned as opposed to the other?
    
    NFS4ERR_RECLAIM_BAD ? is this possible ?
    NFS4ERR_RECLAIM_CONFLICT - claim_prev what we didn't have
    NFS4ERR_GRACE - if use claim_null during grace period that might conflict
    NFS4ERR_NO_GRACE - if try claim_prev outside of grace
    """

    
    def setUp(self):
        NFSSuite.setUp(self)
        self.fh1 = self.fh2 = None
        self.state1 = self.state2 = None
        self.oldleasetime = self.getLeaseTime()

        self.init_connection()
        print

    def tearDown(self):
        # NOTE clean_dir won't work during grace period, nor will close_file
        print "Waiting for grace period to end so we can clean up"
        newleasetime = self.getLeaseTime()
        self.sleep(max(newleasetime, self.oldleasetime))
        self.ncl1.close_file(self.fh1, self.state1, error=None)
        self.ncl2.close_file(self.fh2, self.state2, error=None)
        try: self.clean_dir(self.tmp_dir)
        except SkipException, e:
            pass

    def waitForReboot(self):
        print "Hit ENTER to continue after server is reset"
        sys.stdin.readline()
        print "Continuing with test"
        
    def testRebootValid(self):
        """REBOOT with valid CLAIM_PREVIOUS
        """
        self.fh1, self.state1 = self.ncl.create_confirm(access=OPEN4_SHARE_ACCESS_READ)
        self.waitForReboot()
        
        self.ncl.open_claim_prev(self.fh1, error=[NFS4ERR_STALE_CLIENTID])
        self.init_connection()
        self.fh1, self.state1 = self.ncl.open_claim_prev(self.fh1)


    def testRebootWait(self):
        """REBOOT with late CLAIM_PREVIOUS should return NFS4ERR_NO_GRACE
        """
        self.fh1, self.state1 = self.ncl.create_confirm(access=OPEN4_SHARE_ACCESS_READ)
        self.waitForReboot()
        
        self.ncl.open_claim_prev(self.fh1, error=[NFS4ERR_STALE_CLIENTID])
        self.init_connection()
        self.sleep(90)
        self.ncl.open_claim_prev(self.fh1, error=[NFS4ERR_NO_GRACE])

    def testRebootInvalid(self):
        """REBOOT with invalid CLAIM_PREVIOUS should return NFS4ERR_RECLAIM_CONFLICT or NFS4ERR_RECLAIM_BAD
        """
        self.fh1, self.state1 = self.ncl.create_confirm(access=OPEN4_SHARE_ACCESS_READ)
        self.waitForReboot()
        
        self.init_connection()
        self.fh1, self.state1 = self.ncl.open_claim_prev(self.fh1, OPEN4_SHARE_ACCESS_WRITE, error=[NFS4ERR_RECLAIM_CONFLICT, NFS4ERR_RECLAIM_BAD])

    def testEdge1(self):
        """REBOOT with first edge condition from RFC 3530"""
        
        # Client 1: lock file
        self.fh1, self.state1 = self.ncl1.create_confirm()
        self.ncl1.lock_file(self.fh1, self.state1)

        # Let lease expire
        sleeptime = 3 * self.getLeaseTime() / 2
        self.sleep(sleeptime)

        # Client 2: come in and grab lock
        self.init_connection(ncl=self.ncl2)
        self.fh2, self.state2 = self.ncl2.open_confirm()
        lockid2 = self.ncl2.lock_file(self.fh2, self.state2, type=READ_LT)

        # Client2: now unlock and release the file
        self.ncl2.unlock_file(self.fh2, lockid2, 1)
        self.ncl2.close_file(self.fh2, self.state2)

        self.waitForReboot()

        # Client 1: Reclaim lock (should not work, since #2 has interfered)
        self.ncl1.open_claim_prev(self.fh1, error=[NFS4ERR_STALE_CLIENTID])
        self.init_connection()
        self.fh1, self.state1 = self.ncl.open_claim_prev(self.fh1, error=[NFS4ERR_RECLAIM_CONFLICT, NFS4ERR_RECLAIM_BAD])
        
        
    def testEdge2(self):
        """REBOOT with second edge condition from RFC 3530"""

        # Client 1: lock file
        self.fh1, self.state1 = self.ncl1.create_confirm()
        self.ncl1.lock_file(self.fh1, self.state1)

        self.waitForReboot()

        # Let grace period expire
        sleeptime = 3 * max(self.getLeaseTime(), self.oldleasetime) / 2
        self.sleep(sleeptime)

        # Client 2: come in and grab lock
        self.init_connection(ncl=self.ncl2)
        self.fh2, self.state2 = self.ncl2.open_confirm()
        lockid2 = self.ncl2.lock_file(self.fh2, self.state2, type=READ_LT)

        # Client2: now unlock and release the file
        self.ncl2.unlock_file(self.fh2, lockid2, 1)
        self.ncl2.close_file(self.fh2, self.state2)

        self.waitForReboot()

        # Client 1: Reclaim lock (should not work, since #2 has interfered)
        self.ncl1.open_claim_prev(self.fh1, error=[NFS4ERR_STALE_CLIENTID])
        self.init_connection()
        self.fh1, self.state1 = self.ncl.open_claim_prev(self.fh1, error=[NFS4ERR_RECLAIM_CONFLICT, NFS4ERR_RECLAIM_BAD])
        

    def testRootSquash(self):
        """REBOOT root squash does not work after grace ends?"""
        # Note this assumes we can legally use uid 0...either we are using
        # a secure port or an insecure server
        
        ncl = self.create_client(0, 0, "rootclient")
        self.init_connection(ncl=ncl)

        # See if we are using root squashing
        fh, state = ncl.create_confirm(access=OPEN4_SHARE_ACCESS_READ)
        ops = [ncl.putfh_op(fh)]
        ops.append(ncl.getattr([FATTR4_OWNER]))
        res = ncl.do_ops(ops)
        obj = res.resarray[-1].arm.arm.obj_attributes
        d = nfs4lib.fattr2dict(obj)
        owner = d[FATTR4_OWNER]
        name = owner.split('@',1)[0]
        ncl.close_file(fh, state)
        self.clean_dir(self.tmp_dir)
        if name == "root":
            self.skip("No root squashing detected")
        else:
            print("Detected root squashing root -> %s"%name)

        # Wait for grace period to have *just* expired
        self.waitForReboot()
        self.init_connection(ncl=ncl)
        print "Waiting for grace period to end..."
        while 1:
            fh, state = ncl.create_confirm(access=OPEN4_SHARE_ACCESS_READ,error=[NFS4_OK,NFS4ERR_GRACE])
            if fh: break
            else:
                time.sleep(1)

        # Create a file, and see if root squashing worked
        fh, state = ncl.create_confirm(access=OPEN4_SHARE_ACCESS_READ)
        ops = [ncl.putfh_op(fh)]
        ops.append(ncl.getattr([FATTR4_OWNER]))
        res = ncl.do_ops(ops)
        obj = res.resarray[-1].arm.arm.obj_attributes
        d = nfs4lib.fattr2dict(obj)
        owner = d[FATTR4_OWNER]
        name = owner.split('@',1)[0]
        print "Owner is ", owner
        if name == "root":
            print "WARNING ... must remove file tmp/%s by hand from server test tree"%self.obj_name
            self.fail("Root squashing did not work")
        
        ncl.close_file(fh, state)
